
#ifndef PLS_TASK_MANAGER_H_
#define PLS_TASK_MANAGER_H_

#include <memory>
#include <utility>
#include <array>
#include <atomic>

#include "pls/internal/scheduling/task.h"
#include "pls/internal/scheduling/cont_manager.h"
#include "pls/internal/scheduling/memory_block.h"
#include "pls/internal/data_structures/bounded_ws_deque.h"
#include "pls/internal/data_structures/stamped_integer.h"

namespace pls {
namespace internal {
namespace scheduling {

struct task_handle {
 public:
  task_handle() : task_{nullptr}, task_memory_block_{nullptr} {};
  explicit task_handle(base_task *task) : task_{task},
                                          task_memory_block_{task->get_cont()->get_memory_block()} {};

  base_task *task_;
  // This seems redundant first, but is needed for a race-free steal.
  // It could happen that the task's memory is overwritten and the pointer to it's memory block gets invalid.
  // We can do this more elegantly in the future.
  memory_block *task_memory_block_;
};

/**
 * Handles management of tasks in the system. Each thread has a local task manager,
 * responsible for allocating, freeing and publishing tasks for stealing.
 */
class task_manager {
 public:
  // Publishes a task on the stack, i.e. makes it visible for other threads to steal.
  void publish_task(base_task *task) {
    task_deque_.push_bottom(task_handle{task});
  }

  // Try to pop a local task from this task managers stack.
  bool steal_local_task() {
    return task_deque_.pop_bottom();
  }

  // Try to steal a task from a remote task_manager instance. The stolen task must be stored locally.
  // Returns a pair containing the actual task and if the steal was successful.
  base_task *steal_remote_task(cont_manager &stealing_cont_manager) {
    auto stolen_task_handle = task_deque_.pop_top();
    if (stolen_task_handle) {
      base_task *stolen_task = (*stolen_task_handle).task_;
      auto &atomic_state = (*stolen_task_handle).task_memory_block_->get_state();
      auto &atomic_offered_chain = (*stolen_task_handle).task_memory_block_->get_offered_chain();
      auto offered_chain = stealing_cont_manager.get_node((*stolen_task_handle).task_memory_block_->get_depth());

      auto last_state = atomic_state.load();
      if (last_state.value != memory_block::initialized) {
        return nullptr;
      }

      auto last_offered_chain = atomic_offered_chain.load();
      memory_block::stamped_state loop_state = {last_state.stamp + 1, memory_block::stealing};

      if (atomic_state.compare_exchange_strong(last_state, loop_state)) {
        while (true) {
          if (atomic_offered_chain.compare_exchange_strong(last_offered_chain, offered_chain)) {
            break;
          }

          last_offered_chain = atomic_offered_chain.load();
          last_state = atomic_state.load();
          if (last_state != loop_state) {
            return nullptr;
          }
        }

        if (atomic_state.compare_exchange_strong(loop_state, {loop_state.stamp + 1, memory_block::stolen})) {
          return stolen_task;
        } else {
          return nullptr;
        }
      } else {
        return nullptr;
      }
    } else {
      return nullptr;
    }
  }

  explicit task_manager(data_structures::bounded_ws_deque<task_handle> &task_deque) : task_deque_{task_deque} {}

 private:
  data_structures::bounded_ws_deque<task_handle> &task_deque_;
};

template<size_t NUM_TASKS, size_t MAX_STACK_SIZE>
class static_task_manager {
 public:
  static_task_manager() : task_deque_{}, task_manager_{task_deque_.get_deque()} {};
  task_manager &get_task_manager() { return task_manager_; }

 private:
  data_structures::static_bounded_ws_deque<task_handle, NUM_TASKS> task_deque_;
  task_manager task_manager_;
};

}
}
}
#endif //PLS_TASK_MANAGER_H_
